=== added file '.bzrignore'
--- .bzrignore	1970-01-01 00:00:00 +0000
+++ .bzrignore	2009-01-17 18:36:03 +0000
@@ -0,0 +1,2 @@
+build
+storm.egg-info

=== modified file 'Makefile'
--- Makefile	2008-09-03 00:49:06 +0000
+++ Makefile	2008-11-19 18:22:29 +0000
@@ -21,4 +21,10 @@
 	# Run the tests once with cextensions and once without them.
 	$(TEST_COMMAND) && STORM_CEXTENSIONS=1 $(TEST_COMMAND)
 
+clean:
+	rm -rf build
+	find . -name "*.so" -type f -exec rm -f {} \;
+	find . -name "*.pyc" -type f -exec rm -f {} \;
+	find . -name "*~" -type f -exec rm -f {} \;
+
 .PHONY: all build test

=== modified file 'NEWS'
--- NEWS	2008-11-05 18:47:35 +0000
+++ NEWS	2009-02-20 01:54:08 +0000
@@ -1,4 +1,25 @@
-0.14 (2008-XX-XX)
+0.15 (2009-XX-XX)
+=================
+
+Improvements
+------------
+ - Added support for using Expressions with ResultSet.set().
+ - The default cache size was changed from 100 to 1000 objects.
+ - Added the new GenerationalCache with a faster implementation of
+   the caching API (by Jeroen Vermeulen).  This will likely become
+   the default implementation in the future, so please test it if
+   possible.
+ - A new UUID property type has been added.  It depends on the uuid
+   module introduced in Python 2.5 to represent values.
+
+Bug fixes
+---------
+ - Revert management of distinct, offset and limit by the ResultSet.count()
+   method. It didn't handle some corner cases.
+ - The test suite now passes when run with Python 2.6.
+
+
+0.14 (2009-01-09)
 =================
 
 Improvements
@@ -21,6 +42,11 @@
  - The Reference class now has an __ne__() method, so inequality
    checks against a reference now work in find expressions (bug
    #244768 reported by Stuart Bishop).
+ - Make ResultSet.count() handles the distinct, limit and offset flags, to
+   really reflect the length of the current ResultSet.
+ - The store doesn't iterate on all the alive objects anymore, using instead
+   events that objects can subscribe to. This improves performance drastically
+   when lots of objects are in the cache.
 
 
 0.13 (2008-08-28)

=== modified file 'storm/__init__.py'
--- storm/__init__.py	2008-08-28 14:21:03 +0000
+++ storm/__init__.py	2009-01-09 17:40:25 +0000
@@ -22,7 +22,7 @@
 import os
 
 
-version = "0.13"
+version = "0.14"
 version_info = tuple([int(x) for x in version.split(".")])
 
 

=== modified file 'storm/cache.py'
--- storm/cache.py	2007-12-19 21:55:44 +0000
+++ storm/cache.py	2009-02-16 10:44:31 +0000
@@ -1,3 +1,4 @@
+import itertools
 
 
 class Cache(object):
@@ -9,7 +10,7 @@
     last C{N} C{obj_info}s added to it (where C{N} is the cache size).
     """
 
-    def __init__(self, size=100):
+    def __init__(self, size=1000):
         self._size = size
         self._cache = {} # {obj_info: obj, ...}
         self._order = [] # [obj_info, ...]
@@ -66,3 +67,96 @@
         The most recently added objects come first in the list.
         """
         return list(self._order)
+
+
+class GenerationalCache(object):
+    """Generational replacement for Storm's LRU cache.
+
+    This cache approximates LRU without keeping exact track.  Instead,
+    it keeps a primary dict of "recently used" objects plus a similar,
+    secondary dict of objects used in a previous timeframe.
+
+    When the "most recently used" dict reaches its size limit, it is
+    demoted to secondary dict and a fresh primary dict is set up.  The
+    previous secondary dict is evicted in its entirety.
+
+    Use this to replace the LRU cache for sizes where LRU tracking
+    overhead becomes too large (e.g. 100,000 objects) or the
+    `StupidCache` when it eats up too much memory.
+    """
+
+    def __init__(self, size=1000):
+        """Create a generational cache with the given size limit.
+
+        The size limit applies not to the overall cache, but to the
+        primary one only.  When this reaches the size limit, the real
+        number of cached objects will be somewhere between size and
+        2*size depending on how much overlap there is between the
+        primary and secondary caches.
+        """
+        self._size = size
+        self._new_cache = {}
+        self._old_cache = {}
+
+    def clear(self):
+        """See `storm.store.Cache.clear`.
+
+        Clears both the primary and the secondary caches.
+        """
+        self._new_cache.clear()
+        self._old_cache.clear()
+
+    def _bump_generation(self):
+        """Start a new generation of the cache.
+
+        The primary generation becomes the secondary one, and the old
+        secondary generation is evicted.
+
+        Kids at home: do not try this for yourself.  We are trained
+        professionals working with specially-bred generations.  This
+        would not be an appropriate way of treating older generations
+        of actual people.
+        """
+        self._old_cache, self._new_cache = self._new_cache, self._old_cache
+        self._new_cache.clear()
+
+    def add(self, obj_info):
+        """See `storm.store.Cache.add`."""
+        if self._size != 0 and obj_info not in self._new_cache:
+            if len(self._new_cache) >= self._size:
+                self._bump_generation()
+            self._new_cache[obj_info] = obj_info.get_obj()
+
+    def remove(self, obj_info):
+        """See `storm.store.Cache.remove`."""
+        in_new_cache = self._new_cache.pop(obj_info, None) is not None
+        in_old_cache = self._old_cache.pop(obj_info, None) is not None
+        return in_new_cache or in_old_cache
+
+    def set_size(self, size):
+        """See `storm.store.Cache.set_size`.
+
+        After calling this, the cache may still contain more than `size`
+        objects, but no more than twice that number.
+        """
+        self._size = size
+        cache = itertools.islice(itertools.chain(self._new_cache.iteritems(),
+                                                 self._old_cache.iteritems()),
+                                 0, size)
+        self._new_cache = dict(cache)
+        self._old_cache.clear()
+
+    def get_cached(self):
+        """See `storm.store.Cache.get_cached`.
+
+        The result is a loosely-ordered list.  Any object in the primary
+        generation comes before any object that is only in the secondary
+        generation, but objects within a generation are not ordered and
+        there is no indication of the boundary between the two.
+
+        Objects that are in both the primary and the secondary
+        generation are listed only as part of the primary generation.
+        """
+        cached = self._new_cache.copy()
+        cached.update(self._old_cache)
+        return list(cached)

=== modified file 'storm/cextensions.c'
--- storm/cextensions.c	2008-10-08 14:13:11 +0000
+++ storm/cextensions.c	2009-02-09 16:59:07 +0000
@@ -814,6 +814,7 @@
         Py_INCREF(Undef);
         Py_INCREF(Undef);
         new_value = Undef;
+        Py_DECREF(self->_checkpoint_state);
         self->_checkpoint_state = Undef;
     }
     /* else: */

=== modified file 'storm/properties.py'
--- storm/properties.py	2008-06-26 11:57:22 +0000
+++ storm/properties.py	2009-02-04 06:33:24 +0000
@@ -28,15 +28,15 @@
 from storm.variables import (
     Variable, VariableFactory, BoolVariable, IntVariable, FloatVariable,
     DecimalVariable, RawStrVariable, UnicodeVariable, DateTimeVariable,
-    DateVariable, TimeVariable, TimeDeltaVariable, PickleVariable,
-    ListVariable, EnumVariable)
+    DateVariable, TimeVariable, TimeDeltaVariable, UUIDVariable,
+    PickleVariable, ListVariable, EnumVariable)
 
 
 
 __all__ = ["Property", "SimpleProperty",
            "Bool", "Int", "Float", "Decimal", "RawStr", "Unicode",
-           "DateTime", "Date", "Time", "TimeDelta", "Enum", "Pickle", "List",
-           "PropertyRegistry"]
+           "DateTime", "Date", "Time", "TimeDelta", "UUID", "Enum",
+           "Pickle", "List", "PropertyRegistry"]
 
 
 class Property(object):
@@ -166,6 +166,9 @@
 class TimeDelta(SimpleProperty):
     variable_class = TimeDeltaVariable
 
+class UUID(SimpleProperty):
+    variable_class = UUIDVariable
+
 class Pickle(SimpleProperty):
     variable_class = PickleVariable
 

=== modified file 'storm/store.py'
--- storm/store.py	2008-11-04 16:52:14 +0000
+++ storm/store.py	2009-02-16 10:44:31 +0000
@@ -67,7 +67,7 @@
         """
         @param database: The L{storm.database.Database} instance to use.
         @param cache: The cache to use.  Defaults to a L{Cache} storing up to
-            100 objects at any given time.
+            1000 objects at any given time.
         """
         self._event = EventSystem(self)
         self._connection = database.connect(self._event)
@@ -1283,11 +1283,15 @@
 
         # For now only "Class.attr == var" is supported in args.
         for expr in args:
-            if (not isinstance(expr, Eq) or
-                not isinstance(expr.expr1, Column) or
-                not isinstance(expr.expr2, (Column, Variable))):
+            if not isinstance(expr, Eq):
                 raise FeatureError("Unsupported set expression: %r" %
                                    repr(expr))
+            elif not isinstance(expr.expr1, Column):
+                raise FeatureError("Unsupported left operand in set "
+                                   "expression: %r" % repr(expr.expr1))
+            elif not isinstance(expr.expr2, (Expr, Variable)):
+                raise FeatureError("Unsupported right operand in set "
+                                   "expression: %r" % repr(expr.expr2))
             changes[expr.expr1] = expr.expr2
 
         for key, value in kwargs.items():
@@ -1295,9 +1299,6 @@
             if value is None:
                 changes[column] = None
             elif isinstance(value, Expr):
-                if not isinstance(value, Column):
-                    raise FeatureError("Unsupported set expression: %r" %
-                                       repr(value))
                 changes[column] = value
             else:
                 changes[column] = column.variable_factory(value=value)
@@ -1309,9 +1310,14 @@
         try:
             cached = self.cached()
         except CompileError:
+            # We are iterating through all objects in memory here, so
+            # check if the object type matches to avoid trying to
+            # invalidate a column that does not exist, on an unrelated
+            # object.
             for obj_info in self._store._iter_alive():
-                for column in changes:
-                    obj_info.variables[column].set(AutoReload)
+                if obj_info.cls_info is self._find_spec.default_cls_info:
+                    for column in changes:
+                        obj_info.variables[column].set(AutoReload)
         else:
             changes = changes.items()
             for obj in cached:
@@ -1321,6 +1327,12 @@
                         pass
                     elif isinstance(value, Variable):
                         value = value.get()
+                    elif isinstance(value, Expr):
+                        # If the value is an Expression that means we
+                        # can't compute it by ourselves: we rely on
+                        # the database to compute it, so just set the
+                        # value to AutoReload.
+                        value = AutoReload
                     else:
                         value = variables[value].get()
                     variables[column].set(value)
@@ -1455,7 +1467,7 @@
     def remove(self):
         pass
 
-    def count(self, column=Undef, distinct=False):
+    def count(self, expr=Undef, distinct=False):
         return 0
 
     def max(self, column):

=== modified file 'storm/tracer.py'
--- storm/tracer.py	2008-08-19 13:21:17 +0000
+++ storm/tracer.py	2009-02-04 01:53:02 +0000
@@ -8,7 +8,9 @@
 class DebugTracer(object):
 
     def __init__(self, stream=None):
-        self._stream = stream or sys.stderr
+        if stream is None:
+            stream = sys.stderr
+        self._stream = stream
 
     def connection_raw_execute(self, connection, raw_cursor, statement, params):
         time = datetime.now().isoformat()[11:]

=== modified file 'storm/variables.py'
--- storm/variables.py	2008-11-26 07:28:13 +0000
+++ storm/variables.py	2009-03-04 16:56:58 +0000
@@ -22,6 +22,10 @@
 from decimal import Decimal
 import cPickle as pickle
 import re
+try:
+    import uuid
+except ImportError:
+    uuid = None
 
 from storm.exceptions import NoneError
 from storm import Undef, has_cextensions
@@ -42,6 +46,7 @@
     "TimeVariable",
     "TimeDeltaVariable",
     "EnumVariable",
+    "UUIDVariable",
     "PickleVariable",
     "ListVariable",
 ]
@@ -488,6 +493,24 @@
             return value
 
 
+class UUIDVariable(Variable):
+    __slots__ = ()
+
+    def parse_set(self, value, from_db):
+        assert uuid is not None, "The uuid module was not found."
+        if from_db and isinstance(value, basestring):
+            value = uuid.UUID(value)
+        elif not isinstance(value, uuid.UUID):
+            raise TypeError("Expected UUID, found %r: %r"
+                            % (type(value), value))
+        return value
+
+    def parse_get(self, value, to_db):
+        if to_db:
+            return str(value)
+        return value
+
+
 class EnumVariable(Variable):
     __slots__ = ("_get_map", "_set_map")
 
@@ -538,7 +561,8 @@
         self._event_system = None
 
     def _detect_changes(self, obj_info):
-        if self.get_state() != self._checkpoint_state:
+        if (self._checkpoint_state is not Undef and
+            self.get_state() != self._checkpoint_state):
             self.event.emit("changed", self, None, self._value, False)
 
     def get(self, default=None, to_db=False):

=== modified file 'storm/zope/interfaces.py'
--- storm/zope/interfaces.py	2008-06-26 14:07:36 +0000
+++ storm/zope/interfaces.py	2008-11-27 09:37:56 +0000
@@ -118,6 +118,9 @@
     def cached():
         """Return matching objects from the cache for the current query."""
 
+    def is_empty():
+        """Return true if the result set contains no results."""
+
 
 class ISQLObjectResultSet(Interface):
 

=== modified file 'test'
--- test	2007-09-03 03:11:58 +0000
+++ test	2009-02-05 10:16:40 +0000
@@ -117,7 +117,7 @@
 
     if opts.verbose:
         runner.verbosity = 2
- 
+
     loader = unittest.TestLoader()
 
     unittests, doctests = find_tests(args)
@@ -131,40 +131,41 @@
             self.total_tests += tests
             self.total_failures += failures
             self.total_errors += errors
-            print "(tests=%d, failures=%d, errors=%d)" % \
-                  (tests, failures, errors)
+            sys.stderr.write("(tests=%d, failures=%d, errors=%d)\n" %
+                             (tests, failures, errors))
 
     unittest_summary = Summary()
     doctest_summary = Summary()
 
     if unittests:
-        print "Running unittests..."
+        sys.stderr.write("Running unittests...\n")
         for relpath in unittests:
-            print "[%s]" % relpath
+            sys.stderr.write("[%s]\n" % relpath)
             modpath = relpath.replace(os.path.sep, '.')[:-3]
             module = __import__(modpath, None, None, [""])
             test = loader.loadTestsFromModule(module)
             result = runner.run(test)
             unittest_summary(test.countTestCases(),
                              len(result.failures), len(result.errors))
-            print
+            sys.stderr.write("\n")
 
     if doctests:
-        print "Running doctests..."
+        sys.stderr.write("Running doctests...\n")
         doctest_flags = doctest.ELLIPSIS
         for relpath in doctests:
-            print "[%s]" % relpath
+            sys.stderr.write("[%s]\n" % relpath)
             failures, total = doctest.testfile(relpath,
                                                optionflags=doctest_flags)
             doctest_summary(total, failures, 0)
-            print
+            sys.stderr.write("\n")
 
-    print "Total test cases: %d" % unittest_summary.total_tests
-    print "Total doctests: %d" % doctest_summary.total_tests
-    print "Total failures: %d" % (unittest_summary.total_failures +
-                                  doctest_summary.total_failures)
-    print "Total errors: %d" % (unittest_summary.total_errors +
-                                doctest_summary.total_errors)
+    sys.stderr.write("Total test cases: %d\n" % unittest_summary.total_tests)
+    sys.stderr.write("Total doctests: %d\n" % doctest_summary.total_tests)
+    sys.stderr.write("Total failures: %d\n" %
+                     (unittest_summary.total_failures +
+                      doctest_summary.total_failures))
+    sys.stderr.write("Total errors: %d\n" % (unittest_summary.total_errors +
+                                             doctest_summary.total_errors))
 
     failed = bool(unittest_summary.total_failures or
                   unittest_summary.total_errors or

=== modified file 'tests/cache.py'
--- tests/cache.py	2008-01-08 16:48:55 +0000
+++ tests/cache.py	2009-02-16 10:44:31 +0000
@@ -1,6 +1,6 @@
 from storm.properties import Int
 from storm.info import get_obj_info
-from storm.cache import Cache
+from storm.cache import Cache, GenerationalCache
 
 from tests.helper import TestHelper
 
@@ -21,6 +21,9 @@
         self.hashed = True
         return self.id
 
+    def __lt__(self, other):
+        return self.id < other.id
+
 
 class StubClass(object):
 
@@ -29,28 +32,37 @@
     id = Int(primary=True)
 
 
-class CacheTest(TestHelper):
+class BaseCacheTest(TestHelper):
+
+    Cache = Cache
 
     def setUp(self):
-        super(CacheTest, self).setUp()
+        super(BaseCacheTest, self).setUp()
         self.obj_infos = [StubObjectInfo(i) for i in range(10)]
+        for i in range(len(self.obj_infos)):
+            setattr(self, "obj%d" % (i+1), self.obj_infos[i])
 
     def clear_hashed(self):
         for obj_info in self.obj_infos:
             obj_info.hashed = False
 
+    def test_initially_empty(self):
+        cache = self.Cache()
+        self.assertEqual(cache.get_cached(), [])
+
     def test_add(self):
-        cache = Cache(5)
-        for id in [0, 1, 2]:
-            cache.add(self.obj_infos[id])
-        self.assertEquals([obj_info.id for obj_info in cache.get_cached()],
-                          [2, 1, 0])
+        cache = self.Cache(5)
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+        cache.add(self.obj3)
+        self.assertEquals(sorted(cache.get_cached()),
+                          [self.obj1, self.obj2, self.obj3])
 
     def test_adding_similar_obj_infos(self):
         """If __eq__ is broken, this fails."""
         obj_info1 = get_obj_info(StubClass())
         obj_info2 = get_obj_info(StubClass())
-        cache = Cache(5)
+        cache = self.Cache(5)
         cache.add(obj_info1)
         cache.add(obj_info2)
         cache.add(obj_info2)
@@ -59,42 +71,79 @@
                           [hash(obj_info1), hash(obj_info2)])
 
     def test_remove(self):
-        cache = Cache(5)
-        for id in [0, 1, 2]:
-            cache.add(self.obj_infos[id])
-        cache.remove(self.obj_infos[1])
-        self.assertEquals([obj_info.id for obj_info in cache.get_cached()],
-                          [2, 0])
+        cache = self.Cache(5)
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+        cache.add(self.obj3)
+        cache.remove(self.obj2)
+        self.assertEquals(sorted(cache.get_cached()),
+                          [self.obj1, self.obj3])
 
     def test_add_existing(self):
-        cache = Cache(5)
-        for id in [0, 1, 2, 1]:
-            cache.add(self.obj_infos[id])
-        self.assertEquals([obj_info.id for obj_info in cache.get_cached()],
-                          [1, 2, 0])
-
-    def test_size_and_fifo_behaviour(self):
-        cache = Cache(5)
-        for obj_info in self.obj_infos:
-            cache.add(obj_info)
-        self.assertEquals([obj_info.id for obj_info in cache.get_cached()],
-                          [9, 8, 7, 6, 5])
+        cache = self.Cache(5)
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+        cache.add(self.obj3)
+        cache.add(self.obj2)
+        self.assertEquals(sorted(cache.get_cached()),
+                          [self.obj1, self.obj2, self.obj3])
 
     def test_add_with_size_zero(self):
         """Cache is disabled entirely on add() if size is 0."""
-        cache = Cache(0)
-        obj_info = self.obj_infos[0]
-        cache.add(obj_info)
+        cache = self.Cache(0)
+        cache.add(self.obj1)
         # Ensure that we don't even check if obj_info is in the
         # cache, by testing if it was hashed.  Hopefully, that means
         # we got a faster path.
-        self.assertEquals(obj_info.hashed, False)
+        self.assertEquals(self.obj1.hashed, False)
 
     def test_remove_with_size_zero(self):
         """Cache is disabled entirely on remove() if size is 0."""
-        cache = Cache(0)
-        obj_info = self.obj_infos[0]
-        cache.remove(obj_info)
+        cache = self.Cache(0)
+        cache.remove(self.obj1)
+
+    def test_clear(self):
+        """The clear method empties the cache."""
+        cache = self.Cache(5)
+        for obj_info in self.obj_infos:
+            cache.add(obj_info)
+        cache.clear()
+        self.assertEquals(cache.get_cached(), [])
+
+        # Just an additional check ensuring that any additional structures
+        # which may be used were cleaned properly as well.
+        for obj_info in self.obj_infos:
+            self.assertEquals(cache.remove(obj_info), False)
+
+    def test_set_zero_size(self):
+        """
+        Setting a cache's size to zero clears the cache.
+        """
+        cache = self.Cache()
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+        cache.set_size(0)
+        self.assertEquals(cache.get_cached(), [])
+
+    def test_fit_size(self):
+        """
+        A cache of size n can hold at least n objects.
+        """
+        size = 10
+        cache = self.Cache(size)
+        for value in xrange(size):
+            cache.add(StubObjectInfo(value))
+        self.assertEqual(len(cache.get_cached()), size)
+
+
+class CacheTest(BaseCacheTest):
+
+    def test_size_and_fifo_behaviour(self):
+        cache = Cache(5)
+        for obj_info in self.obj_infos:
+            cache.add(obj_info)
+        self.assertEquals([obj_info.id for obj_info in cache.get_cached()],
+                          [9, 8, 7, 6, 5])
 
     def test_reduce_max_size_to_zero(self):
         """When setting the size to zero, there's an optimization."""
@@ -137,14 +186,197 @@
         self.assertEquals([obj_info.id for obj_info in cache.get_cached()],
                           [5, 4, 3, 2, 1, 0, 9, 8, 7, 6])
 
-    def test_clear(self):
-        """The clear method empties the cache."""
-        cache = Cache(5)
-        for obj_info in self.obj_infos:
-            cache.add(obj_info)
-        cache.clear()
-        self.assertEquals(cache.get_cached(), [])
-
-        # We have two structures. Ensure that both were cleared.
-        for obj_info in self.obj_infos:
-            self.assertEquals(cache.remove(obj_info), False)
+
+class TestGenerationalCache(BaseCacheTest):
+
+    Cache = GenerationalCache
+
+    def setUp(self):
+        super(TestGenerationalCache, self).setUp()
+        self.obj1 = StubObjectInfo(1)
+        self.obj2 = StubObjectInfo(2)
+        self.obj3 = StubObjectInfo(3)
+        self.obj4 = StubObjectInfo(4)
+
+    def test_initially_empty(self):
+        cache = GenerationalCache()
+        self.assertEqual(cache.get_cached(), [])
+
+    def test_cache_one_object(self):
+        cache = GenerationalCache()
+        cache.add(self.obj1)
+        self.assertEqual(cache.get_cached(), [self.obj1])
+
+    def test_cache_multiple_objects(self):
+        cache = GenerationalCache()
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+        self.assertEqual(sorted(cache.get_cached()), [self.obj1, self.obj2])
+
+    def test_clear_cache(self):
+        cache = GenerationalCache()
+        cache.add(self.obj1)
+        cache.clear()
+        self.assertEqual(cache.get_cached(), [])
+
+    def test_clear_cache_clears_the_second_generation(self):
+        cache = GenerationalCache(1)
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+        cache.clear()
+        self.assertEqual(cache.get_cached(), [])
+
+    def test_remove_object(self):
+        cache = GenerationalCache()
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+        cache.add(self.obj3)
+
+        present = cache.remove(self.obj2)
+        self.assertTrue(present)
+        self.assertEqual(sorted(cache.get_cached()), [self.obj1, self.obj3])
+
+    def test_remove_nothing(self):
+        cache = GenerationalCache()
+        cache.add(self.obj1)
+
+        present = cache.remove(self.obj2)
+        self.assertFalse(present)
+        self.assertEqual(cache.get_cached(), [self.obj1])
+
+    def test_size_limit(self):
+        """
+        A cache will never hold more than twice its size in objects.  The
+        generational system is what prevents it from holding exactly the
+        requested number of objects.
+        """
+        size = 10
+        cache = GenerationalCache(size)
+        for value in xrange(5 * size):
+            cache.add(StubObjectInfo(value))
+        self.assertEquals(len(cache.get_cached()), size * 2)
+
+    def test_set_size_smaller_than_current_size(self):
+        """
+        Setting the size to a smaller size than the number of objects
+        currently cached will drop some of the extra content.  Note that
+        because of the generation system, it can actually hold two times
+        the size requested in edge cases.
+        """
+        cache = GenerationalCache(150)
+        for i in range(250):
+            cache.add(StubObjectInfo(i))
+        cache.set_size(100)
+        cached = cache.get_cached()
+        self.assertEquals(len(cached), 100)
+        for obj_info in cache.get_cached():
+            self.assertTrue(obj_info.id >= 100)
+
+    def test_set_size_larger_than_current_size(self):
+        """
+        Setting the cache size to something more than the number of
+        objects in the cache does not affect its current contents,
+        and will merge any elements from the second generation into
+        the first one.
+        """
+        cache = GenerationalCache(1)
+        cache.add(self.obj1) # new=[1]    old=[]
+        cache.add(self.obj2) # new=[2]    old=[1]
+        cache.set_size(2)    # new=[1, 2] old=[]
+        cache.add(self.obj3) # new=[3]    old=[1, 2]
+        self.assertEqual(sorted(cache.get_cached()),
+                         [self.obj1, self.obj2, self.obj3])
+
+    def test_set_size_limit(self):
+        """
+        Setting the size limits the cache's size just like passing an
+        initial size would.
+        """
+        size = 10
+        cache = GenerationalCache(size * 100)
+        cache.set_size(size)
+        for value in xrange(size * 10):
+            cache.add(StubObjectInfo(value))
+        self.assertEquals(len(cache.get_cached()), size * 2)
+
+    def test_two_generations(self):
+        """
+        Inserting more objects than the cache's size causes the cache
+        to contain two generations, each holding up to <size> objects.
+        """
+        cache = GenerationalCache(1)
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+
+        self.assertEqual(sorted(cache.get_cached()), [self.obj1, self.obj2])
+
+    def test_three_generations(self):
+        """
+        If more than 2*<size> objects come along, only 2*<size>
+        objects are retained.
+        """
+        cache = GenerationalCache(1)
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+        cache.add(self.obj3)
+
+        self.assertEqual(sorted(cache.get_cached()), [self.obj2, self.obj3])
+
+    def test_generational_overlap(self):
+        """
+        An object that is both in the primary and the secondary
+        generation is listed only once in the cache's contents.
+        """
+        cache = GenerationalCache(2)
+        cache.add(self.obj1) # new=[1]    old=[]
+        cache.add(self.obj2) # new=[1, 2] old=[]
+        cache.add(self.obj3) # new=[3]    old=[1, 2]
+        cache.add(self.obj1) # new=[3, 1] old=[1, 2]
+
+        self.assertEqual(sorted(cache.get_cached()),
+                         [self.obj1, self.obj2, self.obj3])
+
+    def test_remove_from_overlap(self):
+        """
+        Removing an object from the cache removes it from both its
+        primary and secondary generations.
+        """
+        cache = GenerationalCache(2)
+        cache.add(self.obj1) # new=[1]    old=[]
+        cache.add(self.obj2) # new=[1, 2] old=[]
+        cache.add(self.obj3) # new=[3]    old=[1, 2]
+        cache.add(self.obj1) # new=[3, 1] old=[1, 2]
+
+        present = cache.remove(self.obj1)
+        self.assertTrue(present)
+        self.assertEqual(sorted(cache.get_cached()), [self.obj2, self.obj3])
+
+    def test_evict_oldest(self):
+        """The "oldest" object is the first to be evicted."""
+        cache = GenerationalCache(1)
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+        cache.add(self.obj3)
+
+        self.assertEqual(sorted(cache.get_cached()), [self.obj2, self.obj3])
+
+    def test_evict_LRU(self):
+        """
+        Actually, it's not the oldest but the LRU object that is first
+        to be evicted.  Re-adding the oldest object makes it not be
+        the LRU.
+        """
+        cache = GenerationalCache(1)
+        cache.add(self.obj1)
+        cache.add(self.obj2)
+
+        # This "refreshes" the oldest object in the cache.
+        cache.add(self.obj1)
+
+        cache.add(self.obj3)
+
+        self.assertEqual(sorted(cache.get_cached()), [self.obj1, self.obj3])
+
+
+def test_suite():
+    return unittest.TestLoader().loadTestsFromName(__name__)

=== modified file 'tests/django/stores.py'
--- tests/django/stores.py	2008-08-28 08:19:37 +0000
+++ tests/django/stores.py	2009-02-20 02:04:10 +0000
@@ -26,7 +26,6 @@
     from django.core.exceptions import ImproperlyConfigured
     import transaction
 except ImportError:
-    import traceback; traceback.print_exc()
     have_django_and_transaction = False
 else:
     have_django_and_transaction = True

=== modified file 'tests/infoheritance.txt'
--- tests/infoheritance.txt	2008-09-03 00:06:58 +0000
+++ tests/infoheritance.txt	2009-02-04 02:00:02 +0000
@@ -202,7 +202,7 @@
 >>> store.rollback()
 
 >>> [type(person.info) for person in store.find(Person).order_by(Person.name)]
-[<class 'SecretAgent'>, <class 'Teacher'>]
+[<class '...SecretAgent'>, <class '...Teacher'>]
 >>>
 }}}
 

=== modified file 'tests/properties.py'
--- tests/properties.py	2008-10-06 06:59:51 +0000
+++ tests/properties.py	2009-02-04 06:33:24 +0000
@@ -21,6 +21,10 @@
 from datetime import datetime, date, time, timedelta
 from decimal import Decimal as decimal
 import gc
+try:
+    import uuid
+except ImportError:
+    uuid = None
 
 from storm.exceptions import NoneError, PropertyPathError
 from storm.properties import PropertyPublisherMeta
@@ -550,6 +554,37 @@
 
         self.assertRaises(TypeError, setattr, self.obj, "prop1", object())
 
+    def test_uuid(self):
+        # Skip test if uuid module is not available.
+        if uuid is None:
+            return
+
+        value1 = uuid.UUID("{0609f76b-878f-4546-baf5-c1b135e8de72}")
+        value2 = uuid.UUID("{c9703f9d-0abb-47d7-a793-8f90f1b98d5e}")
+        self.setup(UUID, default=value1, allow_none=False)
+
+        self.assertTrue(isinstance(self.column1, Column))
+        self.assertTrue(isinstance(self.column2, Column))
+        self.assertEquals(self.column1.name, "column1")
+        self.assertEquals(self.column1.table, self.SubClass)
+        self.assertEquals(self.column2.name, "prop2")
+        self.assertEquals(self.column2.table, self.SubClass)
+        self.assertTrue(isinstance(self.variable1, UUIDVariable))
+        self.assertTrue(isinstance(self.variable2, UUIDVariable))
+
+        self.assertEquals(self.obj.prop1, value1)
+        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
+        self.obj.prop2 = None
+        self.assertEquals(self.obj.prop2, None)
+
+        self.obj.prop1 = value1
+        self.assertEquals(self.obj.prop1, value1)
+        self.obj.prop1 = value2
+        self.assertEquals(self.obj.prop1, value2)
+
+        self.assertRaises(TypeError, setattr, self.obj, "prop1",
+                          "{0609f76b-878f-4546-baf5-c1b135e8de72}")
+
     def test_enum(self):
         self.setup(Enum, map={"foo": 1, "bar": 2},
                    default="foo", allow_none=False,

=== modified file 'tests/store/base.py'
--- tests/store/base.py	2008-11-05 18:47:35 +0000
+++ tests/store/base.py	2009-02-19 16:44:33 +0000
@@ -30,7 +30,8 @@
 from storm.properties import PropertyPublisherMeta, Decimal
 from storm.variables import PickleVariable
 from storm.expr import (
-    Asc, Desc, Select, Func, LeftJoin, SQL, Count, Sum, Avg, And, Or, Eq)
+    Asc, Desc, Select, Func, LeftJoin, SQL, Count, Sum, Avg, And, Or, Eq,
+    Lower)
 from storm.variables import Variable, UnicodeVariable, IntVariable
 from storm.info import get_obj_info, ClassAlias
 from storm.exceptions import *
@@ -2320,6 +2321,17 @@
         foo = self.store.get(Foo, 20)
         self.assertEquals(foo.title, "Title 40")
 
+    def test_find_set_with_func_expr(self):
+        self.store.find(Foo, title=u"Title 20").set(title=Lower(u"Title 40"))
+        foo = self.store.get(Foo, 20)
+        self.assertEquals(foo.title, "title 40")
+
+    def test_find_set_equality_with_func_expr(self):
+        self.store.find(Foo, title=u"Title 20").set(
+            Foo.title == Lower(u"Title 40"))
+        foo = self.store.get(Foo, 20)
+        self.assertEquals(foo.title, "title 40")
+
     def test_find_set_column(self):
         self.store.find(Bar, title=u"Title 200").set(foo_id=Bar.id)
         bar = self.store.get(Bar, 200)
@@ -2370,7 +2382,8 @@
     def test_find_set_on_cached_unsupported_python_expr(self):
         foo1 = self.store.get(Foo, 20)
         foo2 = self.store.get(Foo, 30)
-        self.store.find(Foo, Foo.id == Select(SQL("20"))).set(title=u"Title 40")
+        self.store.find(
+            Foo, Foo.id == Select(SQL("20"))).set(title=u"Title 40")
         self.assertEquals(foo1.title, "Title 40")
         self.assertEquals(foo2.title, "Title 10")
 
@@ -2378,9 +2391,14 @@
         result = self.store.find(Foo, title=u"Title 20")
         self.assertRaises(FeatureError, result.set, Foo.title > u"Title 40")
 
-    def test_find_set_expr_unsupported_2(self):
-        result = self.store.find(Foo, title=u"Title 20")
-        self.assertRaises(FeatureError, result.set, Foo.title == Func("func"))
+    def test_find_set_expr_unsupported_without_column(self):
+        result = self.store.find(Foo, title=u"Title 20")
+        self.assertRaises(FeatureError, result.set,
+                          Eq(object(), IntVariable(1)))
+
+    def test_find_set_expr_unsupported_without_expr_or_variable(self):
+        result = self.store.find(Foo, title=u"Title 20")
+        self.assertRaises(FeatureError, result.set, Eq(Foo.id, object()))
 
     def test_find_set_expr_unsupported_autoreloads(self):
         bar1 = self.store.get(Bar, 200)
@@ -2395,6 +2413,47 @@
         self.assertEquals(bar1.title, "Title 400")
         self.assertEquals(bar2.title, "Title 100")
 
+    def test_find_set_expr_unsupported_mixed_autoreloads(self):
+        # For an expression that does not compile (eg:
+        # ResultSet.cached() raises a CompileError), while setting
+        # cached entries' columns to AutoReload, if objects of
+        # different types could be found in the cache then a KeyError
+        # would happen if some object did not have a matching
+        # column. See Bug #328603 for more info.
+        foo1 = self.store.get(Foo, 20)
+        bar1 = self.store.get(Bar, 200)
+        self.store.find(Bar, id=Select(SQL("200"))).set(title=u"Title 400")
+        foo1_vars = get_obj_info(foo1).variables
+        bar1_vars = get_obj_info(bar1).variables
+        self.assertNotEquals(foo1_vars[Foo.title].get_lazy(), AutoReload)
+        self.assertEquals(bar1_vars[Bar.title].get_lazy(), AutoReload)
+        self.assertEquals(bar1_vars[Bar.foo_id].get_lazy(), None)
+        self.assertEquals(foo1.title, "Title 20")
+        self.assertEquals(bar1.title, "Title 400")
+
+    def test_find_set_autoreloads_with_func_expr(self):
+        # In the process of fixing this bug, we've temporarily
+        # introduced another bug: the expression would be called
+        # twice. We've used an expression that increments the value by
+        # one here to see if that case is triggered. In the buggy
+        # bugfix, the value would end up being incremented by two due
+        # to misfiring two updates.
+        foo1 = self.store.get(FooValue, 1)
+        self.assertEquals(foo1.value1, 2)
+        self.store.find(FooValue, id=1).set(value1=SQL("value1 + 1"))
+        foo1_vars = get_obj_info(foo1).variables
+        self.assertEquals(foo1_vars[FooValue.value1].get_lazy(), AutoReload)
+        self.assertEquals(foo1.value1, 3)
+
+    def test_find_set_equality_autoreloads_with_func_expr(self):
+        foo1 = self.store.get(FooValue, 1)
+        self.assertEquals(foo1.value1, 2)
+        self.store.find(FooValue, id=1).set(
+            FooValue.value1 == SQL("value1 + 1"))
+        foo1_vars = get_obj_info(foo1).variables
+        self.assertEquals(foo1_vars[FooValue.value1].get_lazy(), AutoReload)
+        self.assertEquals(foo1.value1, 3)
+
     def test_wb_find_set_checkpoints(self):
         bar = self.store.get(Bar, 200)
         self.store.find(Bar, id=200).set(title=u"Title 400")
@@ -5547,7 +5606,7 @@
     def test_count(self):
         self.assertEquals(self.result.count(), 0)
         self.assertEquals(self.empty.count(), 0)
-        self.assertEquals(self.empty.count(column="abc"), 0)
+        self.assertEquals(self.empty.count(expr="abc"), 0)
         self.assertEquals(self.empty.count(distinct=True), 0)
 
     def test_max(self):

=== modified file 'tests/store/postgres.py'
--- tests/store/postgres.py	2008-11-26 07:56:36 +0000
+++ tests/store/postgres.py	2009-03-04 16:56:58 +0000
@@ -23,12 +23,11 @@
 
 from storm.database import create_database
 from storm.properties import Enum, Int, List
+from storm.info import get_obj_info
 
 from tests.store.base import StoreTest, EmptyResultSetTest, Foo
 from tests.helper import TestHelper
 
-from tests.helper import run_this
-
 
 class Lst1(object):
     __storm_table__ = "lst1"
@@ -189,6 +188,25 @@
         self.store.flush()
         self.assertEquals(foo2.id-foo1.id, 1)
 
+    def test_list_unnecessary_update(self):
+        """
+        Flushing an object with a list variable doesn't create an unnecessary
+        UPDATE statement.
+        """
+        self.store.execute("INSERT INTO lst1 VALUES (1, '{}')", noresult=True)
+
+        lst = self.store.find(Lst1, id=1).one()
+        self.assertTrue(lst)
+        self.store.invalidate()
+
+        lst2 = self.store.find(Lst1, id=1).one()
+        self.assertTrue(lst2)
+        obj_info = get_obj_info(lst2)
+        events = []
+        obj_info.event.hook("changed", lambda *args: events.append(args))
+        self.store.flush()
+        self.assertEquals(events, [])
+
 
 class PostgresEmptyResultSetTest(TestHelper, EmptyResultSetTest):
 

=== modified file 'tests/tracer.py'
--- tests/tracer.py	2008-08-19 13:21:17 +0000
+++ tests/tracer.py	2009-02-04 01:53:02 +0000
@@ -88,7 +88,8 @@
 
     def setUp(self):
         super(DebugTracerTest, self).setUp()
-        self.tracer = DebugTracer()
+        self.stream = self.mocker.mock(file)
+        self.tracer = DebugTracer(self.stream)
 
         datetime_mock = self.mocker.replace("datetime.datetime")
         datetime_mock.now()
@@ -98,13 +99,26 @@
         self.variable = MockVariable("PARAM")
 
     def tearDown(self):
+        del _tracers[:]
         super(DebugTracerTest, self).tearDown()
-        del _tracers[:]
+
+    def test_wb_debug_tracer_uses_stderr_by_default(self):
+        self.mocker.replay()
+
+        tracer = DebugTracer()
+        self.assertEqual(tracer._stream, sys.stderr)
+
+    def test_wb_debug_tracer_uses_first_arg_as_stream(self):
+        self.mocker.replay()
+
+        marker = object()
+        tracer = DebugTracer(marker)
+        self.assertEqual(tracer._stream, marker)
 
     def test_connection_raw_execute(self):
-        stderr = self.mocker.replace("sys.stderr")
-        stderr.write("[04:05:06.000007] EXECUTE: 'STATEMENT', ('PARAM',)\n")
-        stderr.flush()
+        self.stream.write(
+            "[04:05:06.000007] EXECUTE: 'STATEMENT', ('PARAM',)\n")
+        self.stream.flush()
         self.mocker.replay()
 
         connection = "CONNECTION"
@@ -116,9 +130,9 @@
                                            statement, params)
 
     def test_connection_raw_execute_with_non_variable(self):
-        stderr = self.mocker.replace("sys.stderr")
-        stderr.write("[04:05:06.000007] EXECUTE: 'STATEMENT', ('PARAM', 1)\n")
-        stderr.flush()
+        self.stream.write(
+            "[04:05:06.000007] EXECUTE: 'STATEMENT', ('PARAM', 1)\n")
+        self.stream.flush()
         self.mocker.replay()
 
         connection = "CONNECTION"
@@ -130,9 +144,8 @@
                                            statement, params)
 
     def test_connection_raw_execute_error(self):
-        stderr = self.mocker.replace("sys.stderr")
-        stderr.write("[04:05:06.000007] ERROR: ERROR\n")
-        stderr.flush()
+        self.stream.write("[04:05:06.000007] ERROR: ERROR\n")
+        self.stream.flush()
         self.mocker.replay()
 
         connection = "CONNECTION"
@@ -145,9 +158,8 @@
                                                  statement, params, error)
 
     def test_connection_raw_execute_success(self):
-        stderr = self.mocker.replace("sys.stderr")
-        stderr.write("[04:05:06.000007] DONE\n")
-        stderr.flush()
+        self.stream.write("[04:05:06.000007] DONE\n")
+        self.stream.flush()
         self.mocker.replay()
 
         connection = "CONNECTION"
@@ -158,22 +170,6 @@
         self.tracer.connection_raw_execute_success(connection, raw_cursor,
                                                    statement, params)
 
-    def test_custom_stream(self):
-        self.tracer = DebugTracer(sys.stdout)
-
-        stdout = self.mocker.replace("sys.stdout")
-        stdout.write("[04:05:06.000007] EXECUTE: 'STATEMENT', ('PARAM',)\n")
-        stdout.flush()
-        self.mocker.replay()
-
-        connection = "CONNECTION"
-        raw_cursor = "RAW_CURSOR"
-        statement = "STATEMENT"
-        params = [self.variable]
-
-        self.tracer.connection_raw_execute(connection, raw_cursor,
-                                           statement, params)
-
 
 class TimeoutTracerTestBase(TestHelper):
 

=== modified file 'tests/tutorial.txt'
--- tests/tutorial.txt	2008-08-01 17:04:37 +0000
+++ tests/tutorial.txt	2009-02-04 02:00:02 +0000
@@ -83,7 +83,7 @@
 
 {{{#!python
 >>> store.add(joe)
-<Person object at 0x...>
+<...Person object at 0x...>
 
 >>> print "%r, %r" % (joe.id, joe.name)
 None, u'Joe Johnes'
@@ -166,7 +166,7 @@
 >>> mary = Person()
 >>> mary.name = u"Mary Margaret"
 >>> store.add(mary)
-<Person object at 0x...>
+<...Person object at 0x...>
 
 >>> print "%r, %r" % (mary.id, mary.name)
 None, u'Mary Margaret'
@@ -539,7 +539,7 @@
 
 {{{#!python
 >>> store.get(CompanyAccountant, (sweets.id, frank.id))
-<CompanyAccountant object at 0x...>
+<...CompanyAccountant object at 0x...>
 
 }}}
 
@@ -618,7 +618,7 @@
 {{{#!python
 >>> laura = Accountant(u"Laura Montgomery")
 >>> store.add(laura)
-<Accountant ...>
+<...Accountant ...>
 
 >>> subselect = Select(CompanyAccountant.accountant_id, distinct=True)
 >>> result = store.find(Accountant, Not(Accountant.id.is_in(subselect)))
@@ -897,11 +897,11 @@
 >>> list(result)
 [...] EXECUTE: u'SELECT employee.company_id, employee.id, employee.name, "...".company_id, "...".id, "...".name FROM employee, employee AS "..." WHERE employee.company_id = "...".company_id AND employee.id > "...".id', ()
 [...] DONE
-[(<Employee object at ...>, <Employee object at ...>)]
+[(<...Employee object at ...>, <...Employee object at ...>)]
 
 >>> debug(False)
 >>> list(result)
-[(<Employee object at ...>, <Employee object at ...>)]
+[(<...Employee object at ...>, <...Employee object at ...>)]
 
 }}}
 

=== modified file 'tests/variables.py'
--- tests/variables.py	2008-11-26 07:37:50 +0000
+++ tests/variables.py	2009-03-04 16:56:58 +0000
@@ -21,6 +21,12 @@
 from datetime import datetime, date, time, timedelta
 from decimal import Decimal
 import cPickle as pickle
+import gc
+import weakref
+try:
+    import uuid
+except ImportError:
+    uuid = None
 
 from storm.exceptions import NoneError
 from storm.variables import *
@@ -91,6 +97,18 @@
         variable.set(marker, from_db=True)
         self.assertEquals(variable.sets, [(marker, False), (marker, True)])
 
+    def test_set_leak(self):
+        """When a variable is checkpointed, the value must not leak."""
+        variable = Variable()
+        m = Marker()
+        m_ref = weakref.ref(m)
+        variable.set(m)
+        variable.checkpoint()
+        variable.set(LazyValue())
+        del m
+        gc.collect()
+        self.assertIdentical(m_ref(), None)
+
     def test_get(self):
         variable = CustomVariable()
         variable.set(marker)
@@ -763,6 +781,47 @@
             self.fail("ValueError not raised")
 
 
+class UUIDVariableTest(TestHelper):
+
+    def is_supported(self):
+        return uuid is not None
+
+    def test_get_set(self):
+        value = uuid.UUID("{0609f76b-878f-4546-baf5-c1b135e8de72}")
+
+        variable = UUIDVariable()
+
+        variable.set(value)
+        self.assertEquals(variable.get(), value)
+        self.assertEquals(
+            variable.get(to_db=True), "0609f76b-878f-4546-baf5-c1b135e8de72")
+
+        self.assertRaises(TypeError, variable.set, marker)
+        self.assertRaises(TypeError, variable.set,
+                          "0609f76b-878f-4546-baf5-c1b135e8de72")
+        self.assertRaises(TypeError, variable.set,
+                          u"0609f76b-878f-4546-baf5-c1b135e8de72")
+
+    def test_get_set_from_database(self):
+        value = uuid.UUID("{0609f76b-878f-4546-baf5-c1b135e8de72}")
+
+        variable = UUIDVariable()
+
+        # Strings and UUID objects are accepted from the database.
+        variable.set(value, from_db=True)
+        self.assertEquals(variable.get(), value)
+        variable.set("0609f76b-878f-4546-baf5-c1b135e8de72", from_db=True)
+        self.assertEquals(variable.get(), value)
+        variable.set(u"0609f76b-878f-4546-baf5-c1b135e8de72", from_db=True)
+        self.assertEquals(variable.get(), value)
+
+        # Some other representations for UUID values.
+        variable.set("{0609f76b-878f-4546-baf5-c1b135e8de72}", from_db=True)
+        self.assertEquals(variable.get(), value)
+        variable.set("0609f76b878f4546baf5c1b135e8de72", from_db=True)
+        self.assertEquals(variable.get(), value)
+
+
 class PickleVariableTest(TestHelper):
 
     def test_get_set(self):

